from webob.request import BaseRequest, environ_from_url#, environ_add_POST
from webob.response import Response as WebobResponse, iter_close
from mext.dispatch import Adapter
from mext.uri import AbsoluteUri


__all__ = [
    'Request', 'Response', 'OverrideResponse',
    'WSGI', 'wapp', 'wapp_mw',
]

def _set_attrs(ob, kw):
    cls = ob.__class__
    for name, value in kw.iteritems():
        if not hasattr(cls, name):
            raise TypeError("Unexpected keyword: %s=%r" % (name, value))
        setattr(ob, name, value)


class Request(BaseRequest):
    _charset_cache = None
    #charset = property(staticmethod(lambda: 'UTF-8'))
    charset = property(lambda self: 'UTF-8')
    uri = property(lambda self: AbsoluteUri(self.url))
    app_uri = property(lambda self: AbsoluteUri(self.application_url))
    appdir_uri = property(lambda self: AbsoluteUri(self.application_url).as_dir())

    bGET, uGET = BaseRequest.str_GET, BaseRequest.GET
    bPOST, uPOST = BaseRequest.str_POST, BaseRequest.POST
    bparams, uparams = BaseRequest.str_params, BaseRequest.params
    bcookies, ucookies = BaseRequest.str_cookies, BaseRequest.cookies

    def __init__(self, env='/', **kw):
        if isinstance(env, str):
            env = environ_from_url(env)
        self.environ = env
        if kw:
            _set_attrs(self, kw)

    def get_response(self, app):
        status, headers, app_iter = self.call_application(app)
        return Response(status=status, headerlist=list(headers), app_iter=app_iter)
    #__rshift__ = get_response # resp = req>>app
    __rlshift__ = get_response # resp = app<<req


class Response(WebobResponse):
    adapt = Adapter('__webob__', subadapt=False)
    conditional_response = property()

    @classmethod
    def make(cls, ob, **kw):
        r = cls.adapt(ob)
        for k, v in kw.items():
            assert hasattr(cls, k)
            setattr(r, k, v)
        return r

    _status = '200 OK'
    _environ = _request = None
    _body = _app_iter = _headers = None

    def __init__(self, body=None, status=None, headerlist=None, app_iter=None, content_type=None, charset=None, **kw):
        if status is not None:
            self.status = status

        if headerlist is not None:
            assert content_type is charset is None
            self._headerlist = headerlist
        elif content_type is None:
            assert charset is None
            self._headerlist = [('Content-Type', 'text/html; charset=UTF-8')]
            charset = 'UTF-8'
        else:
            if charset:
                content_type += '; charset=' + charset
            elif 'charset=' in content_type:
                pass
                # we can detect charset for ubody, but that's a very strange case
                #charset = content_type.rsplit('charset=', 1)[1]
            elif charset is not None:
                # pass charset='' or charset=False (anything false but not None)
                # to disable automatic utf-8 charset for text and xml responses
                pass
            elif (content_type.startswith('text/')
                or (content_type.startswith('application/') and content_type.endswith('xml'))
            ):
                # assert charset is None
                content_type += '; charset=UTF-8'
                charset = 'UTF-8'
            self._headerlist = [('Content-Type', content_type)]

        if app_iter is not None:
            assert body is None
            self._app_iter = app_iter
        else:
            body = body or ''
            if isinstance(body, unicode):
                assert charset, "Got unicode body but no charset"
                body = body.encode(charset)
            self._body = body
            assert not headerlist or ('Content-Length' not in self.headers)
            self._headerlist.append(('Content-Length', str(len(body))))
        if kw:
            _set_attrs(self, kw)

    def __call__(self, environ, start_response):
        start_response(self.status, self.headerlist)
        if environ['REQUEST_METHOD'] == 'HEAD':
            iter_close(self.app_iter)
            return ()
        return self.app_iter

    def flat_app_iter(self):
        r = self._app_iter
        if r is None:
            return [self._body]
        if not isinstance(r, (list, tuple)):
            r = self._app_iter = list(r)
        return r


class OverrideResponse(BaseException):
    def __init__(self, app, comment=None):
        self.app = app
        self.comment = comment

#@@
#Response.adapt[object] = lambda ob: MimeTypedData.adapt(ob).__webob__()
#@@ .chain_adapter
## @register_response_body_cls('xmlgen.XmlDocument')
## def transform_xmldoc(doc):
##     return dict(body=str(doc), charset=doc.encoding, content_type=doc.mime_type)
##
## @register_response_body_cls('lxml.html.HtmlElement')
## def transform_htmldoc(doc, charset='UTF-8'):
##     from lxml.html import tostring
##     return dict(body=tostring(elem, encoding=charset), charset=charset)

WSGI = Adapter('__wsgi__')


@WSGI.register(object)
def adapt_wsgi(ob):
    if hasattr(ob, '__call__'):
        return ob
    else:
        return Response.adapt(ob)



class wapp(object):
    __slots__ = ('func', '_func_has_get')
    def __init__(self, func):
        self.func = func
        self._func_has_get = hasattr(func, '__get__')

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.func)

    def __get__(self, ob, type):
        if self._func_has_get:
            return self.__class__(self.func.__get__(ob, type))
        else:
            return self

    def __call__(self, env, start):
        req = Request(env)
        try:
            app = WSGI(self.func(req))
        except OverrideResponse, exc:
            app = exc.app
        return app(env, start)


class wapp_mw(object):
    __slots__ = ('mwfunc', 'next_app', 'args', 'kw')
    def __new__(cls, *args, **kw):
        if len(args) == 1 and not kw:
            mwfunc = args[0]
            factory = lambda *args, **kw: cls(mwfunc, *args, **kw)
            factory.__name__ = 'wapp_mw(%r)' % mwfunc
            return factory
        return object.__new__(cls)

    def __init__(self, mwfunc, next_app, *args, **kw):
        self.mwfunc = mwfunc
        self.next_app = next_app
        self.args = args
        self.kw = kw

    def __repr__(self):
        kwrepr = ''.join(
            map(', %r'.__mod__, self.args)
            + map(', %s=%r'.__mod__, self.kw.items())
        )
        return '%s(%r, %r%s)' % (self.__class__.__name__, self.mwfunc, self.next_app, kwrepr)

    @wapp
    def __call__(self, req):
        return self.mwfunc(req, self.next_app, *self.args, **self.kw)

