from __future__ import with_statement

import sys, os, traceback, pprint, threading
import itertools, types, time, re
from cStringIO import StringIO

from pasteob import *
from pasteob.fileapp import PkgResourcesApp
from pasteob.debug.show_source import show_module_source_app
from pasteob.debug.collect import Traceback
from pasteob.debug.report import WebReport
from pasteob.debug.utils import *

from mext.htmlgen import *


# TODO
# Contextual state in extra info
# save contextual state for eval



__all__ = ['EvalDebugger']



class EvalDebugger(object):
    debug_prefix = '__debug__'

    def __init__(self, app, debug_app=None):
        self.app = app
        if debug_app is None:
            debug_app = DebugApp()
        self.debug_app = debug_app

    @wapp
    def __call__(self, req):
        assert not req.environ['wsgi.multiprocess']
        if req.path_info_peek() == self.debug_prefix:
            req.path_info_pop()
            return self.debug_app
        else:
            return self.wrapped_app

    @wapp
    def wrapped_app(self, req):
        # get base_path before script_path could have been modified
        base_path = '%s/%s' % (req.application_url, self.debug_prefix)
        try:
            r = self.app << req
            r.flat_app_iter()
            return r
        except:
            id = new_id()
            assert id not in self.debug_app.traces
            tb = Traceback.collect()
            rep = WebReport(req, tb)
            #tb.extra_data.update(make_extra_data(req, self.app))

            trace_app = TraceApp(id, rep, base_path)
            req.environ['wsgi.errors'].write('Debug at: %s\n' % trace_app.view_uri)

            self.debug_app.traces[id] = trace_app
            #req.environ['dont-clear'] = True
            return trace_app.GET




class DebugApp(Autodispatch):
    def __init__(self):
        self.traces = {}

    dispatch_media = PkgResourcesApp('pasteob.debug.eval', 'media')
    dispatch_src = show_module_source_app

    @wapp
    def GET_summary(self, req):
        """
            Returns a JSON-format summary of all the cached
            exception reports
        """
        data = []
        items = self.traces.values()
        items.sort(key=lambda v: v.created)
        data = [item.json() for item in items]
        return Response(unicode_repr_safe(data),
            content_type='text/x-json'
            #content_type='text/plain'
        )


    @wapp
    def __default__(self, req):
        id = req.path_info_pop()
        if id not in self.traces:
            return ErrorResponse(404,
                "Traceback by id %s does not exist (maybe "
                "the server has been restarted?)" % id)
        return self.traces[id]






exec_lock = threading.Lock()

def ctx_exec(stmt, locals, globals):
    with exec_lock:
        out = StringIO()
        sys.stdout, save_stdout = out, sys.stdout
        try:
            try:
                code = compile(stmt, '<web>', "single", 0, 1)
                exec code in globals, locals
            except KeyboardInterrupt:
                raise
            except:
                traceback.print_exc(file=out)
        finally:
            sys.stdout = save_stdout
        return out.getvalue()




new_id = lambda: os.urandom(4).encode('hex')

class TraceApp(Autodispatch):
    def __init__(self, trace_id, rep, base_path):
        self.id = trace_id
        self.rep = rep
        self.base_path = base_path
        self.view_uri = '%s/%s/' % (base_path, trace_id)
        self.created = time.time()
        self.css_link = LINK_css(self.base_path+'/media/traceback.css', media='screen')

    @wapp
    def GET(self, req): #@@
        tb_html = EvalHTMLFormatter(self.rep, self.base_path, self.id).do_format()
        traceback_body = DIV('short_traceback')[tb_html]

        title = self.rep.tb_last_line
        r = Response.make(
            HtmlDocument([
                TITLE[title],
                self.css_link,
                H1[title],
                self.tmpl_links(),
                self.tmpl_js_setup(),
                traceback_body,
            ]),
            status_int=500
        )
        r.headers['X-Debug-URL'] = self.view_uri
        return r


    def tmpl_links(self):
        A_trace = lambda text, suffix='': A('%s/%s/%s' % (self.base_path, self.id, suffix))[text]
        return DIV[
            A(self.rep.req.url)['original'], ' / ',
            A_trace('permalink'), ' / ',
            #A_trace('extra data', 'extra_data'), ' / ',
            A_trace('plain traceback', 'plain_tb'),
        ], BR

    def tmpl_js_setup(self):
        return [
            NOSCRIPT[DIV(cls='warn')["Enable JS to use interactive debugging."]],
            DIV('error-area', cls='error', style='display: none')[
                BUTTON('clear this', "$('#error-container').empty();$('#error-area').hide()"),
                DIV('error-container')[''],
            ],
            SCRIPT_js('debug_base="%s/%s/"' % (self.base_path, self.id)),
            LINK_js('http://ajax.googleapis.com/ajax/libs/jquery/1.4.2/jquery.min.js'),
            # google cdn has lower latency in my tests
            #LINK_js('http://code.jquery.com/jquery-1.4.2.min.js'),
            LINK_js(self.base_path+'/media/debug.js'),
        ]

##     @wapp
##     def GET_extra_data(self, req): #@@ syntax h-l etc
##         traceback_body, extra_data = self.format_eval_html()
##         return HtmlDocument(extra_data)
##



    @wapp
    def GET_plain_tb(self, req): #@@ syntax h-l etc
        text = self.rep.tb_text
        return HtmlDocument(
            TEXTAREA(rows=text.count('\n'), cols=60, style='width: 100%')[text]
        )



    @wapp
    def GET_show_frame(self, req):
        tbid = int(req.params['tbid'])
        frame = self.rep.tb.frames[tbid]
        vars = frame.locals
        local_vars = make_table(vars) if vars else 'No local vars'
        return HtmlDocument([make_input_form(tbid), local_vars])

    @wapp
    def POST_exec_input(self, req):
        input = req.params.get('input')
        if not input.strip():
            return Response()
        input = input.rstrip()
        tbid = int(req.params['tbid'])
        frame = self.rep.tb.frames[tbid]
        output = ctx_exec(input,
            frame.locals,
            frame.globals #@@
        )
        r = NOTAG
        for i, line in enumerate(input.splitlines()):
            prompt = '>>> ' if (i==0) else '... '
            r += SPAN('prompt')[prompt], line, '\n'
        r += output
        return Response(as_xhtml(r))



    def json(self):
        """Return the JSON-able representation of this object"""
        created = self.rep.tb.timestamp
        return {
            'uri': self.view_uri,
            'created': created.strftime('%c'),
            'created_timestamp': created,
            'exception_type': unicode_safe(self.rep.tb.exc_type),
            'exception': unicode_safe(self.rep.tb.exc),
        }





class EvalHTMLFormatter(object):
    def __init__(self, rep, base_path, id):
        self.rep = rep
        self.base_path = base_path
        self.img_plusminus = [
            IMG("%s/media/%s.jpg" % (self.base_path, name), cls=name, border=0, width=9, height=9)
            for name in ('minus', 'plus')
        ]
        self.id = id

    def do_format(self):
        #@@lines += self.extras_html(data_by_importance['important'])
        tb = DIV('traceback')
        for i, frame in enumerate(self.rep.tb.frames):
            fdiv = DIV(cls='frame')[self.format_source_line(i, frame)]
            source = frame.get_source_line()
            if source:
                onclick_js = "$(this).hide().closest('.frame').find('.src-long').show()"
                fdiv += DIV(cls='src-short', onclick=onclick_js)[PRE_flex(source.strip())]
                fdiv += DIV(cls='src-long')[PRE_flex(frame.get_source_line(3))]
            tb += fdiv
        tb += BOLD[self.rep.tb_last_line]

        #for name in 'normal', 'supplemental':
        #    lines += self.extras_html(data_by_importance[name])
        #extra_data = self.extras_html(data_by_importance['extra'])
        return tb



    def format_source_line(self, i, frame):
        link = '%s/src/%s#code-%s' % (self.base_path, frame.module_name, frame.line_no)
        return DIV[
            'Module ', A(link, target="_blank")[BOLD[frame.module_name], ':%s' % frame.line_no],
            ' in ', CODE(cls='func')[frame.func_name or '?'], ' ',
            A('#', cls='show-locals', tbid=i, onclick="return show_frame(this)")[
                self.img_plusminus, ' debug'
            ]
        ]



def _html_repr(v):
    try:
        r = pprint.pformat(v)
    except Exception, e:
        return SPAN('error')[u'Error: %s' % e]
    if len(r) < 150:
        return PRE_flex(r)
    extra = r[100:]
    long_id = 'extra-' + new_id()
    onclick_js = '$(this).hide(); $("#%s").show(); return false' % long_id
    r = PRE_flex(r[:100])
    r += SPAN('more', onclick=onclick_js)['%d more chars' % len(extra)]
    r += SPAN(id=long_id, style='display: none')[extra]
    return r




def make_table(data):
    r = TABLE('repr')
    even = False
    for name, value in sorted(data.items()):
        r += TR(cls='even' if even else 'odd')[TH[name], TD[_html_repr(value)]]
        even = not even
    return r




def make_input_form(tbid):
    onsubmit_js = "return submit_input('%r')" % tbid
    return FORM(onsubmit=onsubmit_js, id='form-%d' % tbid)[
        PRE(cls='output', style='display: none')[''],
        INPUT('text', 'input', cls='exec', autocomplete='off'), BR,
        BUTTON("Execute", onsubmit_js),
        BUTTON("Expand", "return expand_input(this)"),
    ]


##     def format_extra_data(self, importance, title, value):
##         if isinstance(value, str):
##             s = pretty_string_repr(value)
##             if '\n' in s:
##                 return [title, ':', BR, PRE[s]]
##             else:
##                 return [title, ':', TT[s]]
##         elif isinstance(value, dict):
##             return H3[title], zebra_table(value)
##         else:
##             return [title, ': ', PRE_flex[pprint.pformat(vrepr)]]
##

##
## def zebra_table(rows, table_class="variables"):
##     if isinstance(rows, dict):
##         rows = rows.items()
##         rows.sort()
##     r = TABLE(table_class)
##     #r += TR('header')[TH(colspan=2)[title]]
##     odd = False
##     for name, value in rows:
##         try:
##             value = repr(value)
##         except Exception, e:
##             value = 'Cannot print: %s' % e
##         odd = not odd
##         r += TR('odd' if odd else 'even')[
##             TD[name],
##             TD[CODE[truncate(value)]],
##         ]
##     return r
##
##
##
##


def pretty_string_repr(s):
    """
        Formats the string as a triple-quoted string when it contains
        newlines.
    """
    if '\n' in s:
        s = repr(s)
        s = s[0]*3 + s[1:-1] + s[-1]*3
        s = s.replace('\\n', '\n')
        return s
    else:
        return repr(s)


def make_extra_data(req, app):
    cgi_vars = {}
    wsgi_vars = {}
    for name, value in req.environ.items():
        if name.isupper():
            if value:
                cgi_vars[name] = value
        elif name not in hide_vars:
            wsgi_vars[name] = value
    wsgi_vars['application'] = app
    return {
        ('extra', 'Request'): [req.__str__(True)],
        ('extra', 'CGI Variables'): [cgi_vars],
        ('extra', 'WSGI Variables'): [wsgi_vars]
    }

hide_vars = set([
    'paste.config', 'wsgi.errors', 'wsgi.input',
    #'wsgi.multithread', 'wsgi.multiprocess', 'wsgi.run_once',
    #'wsgi.version', 'wsgi.url_scheme'
])


