import sys, os, socket, traceback, select, signal
from datetime import datetime

from webob.response import iter_close
from pasteob import *
from pasteob.server.request import parse_request, CloseConnection#, SockRFile

from mext.threads import Thread, Flag, Queue
from mext.log import Logger
from mext.context import State
from mext.uri import Uri

log = Logger(__name__, min_level='INFO')

__all__ = ['WsgiServer']


class ServerExit(SystemExit):
    pass

is_win = (os.name == 'nt')

# @@ support chunked encoding

def new_ipv4_socket():
    return socket.socket()
    #return socket._realsocket(socket.AF_INET, socket.SOCK_STREAM)

_new_id = iter(xrange(1, sys.maxint)).next

class WsgiServer(object):
    request_queue_size = 20
    allow_reuse_address = False
    socket_timeout = None
    handle_signals = True # will not work unless running in main thread
    worker_count = 3

    #
    # initial setup
    #

    def __init__(self, app, host=None, port=8080, state=None):
        self.app = app
        self.host = host or '127.0.0.1'
        self.port = port
        self.state = state or State.clone() # state to clone for worker threads
        http_host = self.host
        if self.port != 80:
            http_host += ':%s' % self.port
        self.uri = Uri('http://%s/' % http_host)
        self._serving = False
        self.started = Flag()
        self.stopped = Flag()

    def __enter__(self):
        Thread(self.serve).start()
        self.started.wait()
        return self

    def __exit__(self, *args):
        self.shutdown()


    #
    # Eventloop initialization
    #

    def _eventloop_setup(self):
        self.idle_conns = [] # a list of connections in keep-alive state
        self.create_socket()
        self.create_wakeup_socket()
        self.req_queue = Queue() # a queue of connections with requests pending
        self.start_workers()
        if self.handle_signals and Thread.get().is_main:
            self.setup_signals()
        self._serving = True
        self.started.set()

    def create_socket(self):
        self.socket = new_ipv4_socket()
        if self.allow_reuse_address:
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((self.host, self.port))
        self.socket.listen(self.request_queue_size)
        #self.host, self.port = self.socket.getsockname()
        #self.server_name = socket.getfqdn(self.host)

    def create_wakeup_socket(self):
        sock = new_ipv4_socket()
        sock.bind(('127.0.0.1', 0))
        sock.listen(1)
        addr = sock.getsockname()
        wkw = new_ipv4_socket()
        wkw.connect(addr)
        wkr, _ = sock.accept()
        self._wakeup_w = wkw
        self._wakeup_r = wkr

    def start_workers(self):
        self.workers = []
        for i in xrange(self.worker_count):
            self._start_worker()

    def _start_worker(self):
        th = Thread(self.worker_loop)
        th.name = 'worker-loop-%d' % _new_id()
        th.start_daemon()
        self.workers.append(th)


    #
    # Shutdown process
    #

    def shutdown(self, timeout=None):
        """
            Request the server to shut down. Can be called from any thread.
            If called from a non-worker thread will wait until the server has completely stopped.
        """
        if self._serving:
            log.info("Requesting shutdown")
            self._serving = False
            self.wakeup()
            if Thread.get() not in self.workers:
                self.stopped.wait(timeout)

    def _eventloop_cleanup(self):
        log.info("Shutting down")
        self._serving = False
        if self.handle_signals and Thread.get().is_main:
            self.restore_signals()
        log.debug("Closing idle sockets")
        self.socket = None
        del self.idle_conns[:]
        log.debug("Requesting workers to stop")
        for i in xrange(len(self.workers)):
            self.req_queue.put(None) # quit request
        for th in self.workers[:]:
            th.join()
        log.debug("Closing remaining sockets")
        self._wakeup_r = self._wakeup_w = None
        del self.idle_conns[:]
        log.debug("Shutdown successful")
        self.stopped.set()



    #
    # eventloop
    #

    def wakeup(self):
        #log.debug("Request wakeup")
        self._wakeup_w.send('x')

    def serve(self, poll_interval=None):
        assert not self._serving
        log.info("Serving on %s", self.uri)
        if poll_interval is None:
            poll_interval = 0.5 if is_win else 5
        self._eventloop_setup()
        try:
            self._eventloop_run(poll_interval)
        finally:
            self._eventloop_cleanup()

    def _eventloop_run(self, poll_interval):
        while self._serving:
            srv = self.socket
            wkr = self._wakeup_r
            r = [srv, wkr] + self.idle_conns
            try:
                r, w, e = select.select(r, [], [], poll_interval)
            except select.error, exc:
                if exc[0] == 4: # signal
                    continue
                raise
            for sock in r:
                if sock is wkr:
                    #log.debug("Woke up")
                    wkr.recv(100) # consume all wakeup requests
                elif sock is srv:
                    try:
                        conn, caddr = srv.accept()
                        log.debug("New connection from %s:%s", *caddr)
                        conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                        if self.socket_timeout:
                            conn.settimeout(self.socket_timeout)
                    except socket.error:
                        continue
                    self.process_request(conn)
                else:
                    self.idle_conns.remove(sock)
                    #log.debug("New request")
                    self.process_request(sock)



    #
    # Worker threads
    #

    def process_request(self, conn):
        self.req_queue.put(conn)

    def worker_loop(self):
        try:
            #thlog = Logger(Thread.get().name, parent=log)
            while True:
                #thlog.debug("Waiting for request")
                conn = self.req_queue.get()
                if conn is None: # quit
                    break
                #thlog.debug("Got request from %s:%d", *conn.getpeername())
                self.handle_req(conn)
                #thlog.debug("Done processing request")
        finally:
            th = Thread.get_current()
            self.workers.remove(th)

    def handle_req(self, conn):
        try:
            #caddr = conn.getpeername()
            self._handle_req(conn)
        except socket.error, exc:
            pass
        except CloseConnection:
            log.debug("Close connection to %s:%d", *conn.getpeername())
            conn.close()
            #SockRFile.delete_from(conn)
        except:
            log.error("Unexpected worker error", exc_info=True)
        else:
            self.idle_conns.append(conn)
            self.wakeup() # make sure conn participates in select()

    def _handle_req(self, conn):
        caddr = conn.getpeername()
        env = resp = None
        http10 = True
        with self.state.clone():
            try:
                env = parse_request(conn)
                if not isinstance(env, dict):
                    # env is an ErrorResponse instance
                    send_response(env.text_response, conn)
                    raise CloseConnection
                env.update({
                    'pasteob.server': self,
                    'SERVER_NAME': self.host,
                    'SERVER_PORT': self.port,
                })
                http10 = (env['SERVER_PROTOCOL'] == 'HTTP/1.0')
                # we use env copy so that if apps change the env (for example method)
                #log.debug(str(Request(env).uri))
                resp = self.app << Request(env.copy())
            except CloseConnection:
                raise
            except socket.error:
                raise CloseConnection
            except:
                log.error("Error handling request from %s", caddr[0], exc_info=True)
                disconnect = True
                resp = ErrorResponse(500) << Request(env.copy())
            else:
                disconnect = decide_disconnect(env, resp)
            # if the server is shutting down, try to close the connections gracefully
            resp = filter_response(env, resp)
            if not self._serving:
                disconnect = True
            if http10 and not disconnect:
                resp.headers['connection'] = 'keep-alive'
            elif not http10 and disconnect:
                resp.headers['connection'] = 'close'
            send_response(resp, conn)
        # dont-clear is a flag set by pasteob.debug.eval so the data is there
        # for the debugging
        #if req and 'dont-clear' not in req.environ:
        #    req.environ.clear() # remove cycles created by ReqAddons
        if disconnect:
            raise CloseConnection



    #
    # Signal handling
    #

    def setup_signals(self):
        self._set_signals(self.on_term)

    def restore_signals(self):
        self._set_signals(signal.SIG_DFL)

    def _set_signals(self, handler):
        if is_win:
            signal.signal(signal.SIGBREAK, handler)
        signal.signal(signal.SIGTERM, handler)

    def on_term(self, sig, fr):
        log.info("Got TERM signal")
        self.shutdown(0)





################
################
################


def filter_response(env, resp):
    status_int = resp.status_int
    if (env['REQUEST_METHOD'] == 'HEAD'
        or status_int < 200
        or status_int in (204, 304)
    ):
        iter_close(resp.app_iter)
        resp.app_iter = ()
    return resp


def send_response(resp, conn):
    resp.server = 'PasteOb/0.7'
    if not resp.date:
        resp.date = datetime.now()
    if resp.content_length is None:
        resp.headers['connection'] = 'close'

    headers = ['HTTP/1.1 ' + resp.status]
    headers += map('%s: %s'.__mod__, resp.headerlist)
    headers += ('', '')
    conn.sendall('\r\n'.join(headers))
    try:
        for chunk in resp.app_iter:
            conn.sendall(chunk)
    finally:
        iter_close(resp.app_iter)

def decide_disconnect(env, resp):
    req_conn = env.get('HTTP_CONNECTION', '').lower()
    resp_conn = resp.headers.get('connection', '').lower()
    if req_conn == 'close' or resp_conn == 'close':
        return True
    elif resp.status_int == 413: # request entity too large
        return True
    elif (resp.content_length is None
        and env['REQUEST_METHOD'] != 'HEAD'
        and resp.status_int >= 200
        and resp.status_int not in (204, 304)
    ):
        #@@ if http/1.1, convert response to chunked
        return True
    elif req_conn == 'keep-alive':
        return False
    else:
        return (env['SERVER_PROTOCOL'] == 'HTTP/1.0')






if __name__ == '__main__':
    log.min_level = 'debug'
    from pprint import pformat
    @wapp
    def dump_req(req):
        #print req
        if req.query_string == 'shutdown':
            req.environ['pasteob.server'].shutdown()
        return Response(str(req)+'\n'+pformat(req.environ), content_type='text/plain')
    WsgiServer(dump_req).serve()

